package org.coderu.core.impl.explorer.rules;

import static com.google.common.base.Predicates.*;
import static org.coderu.common.utils.CollectionUtils.*;
import static org.coderu.core.impl.common.types.PackageName.*;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.coderu.common.utils.CollectionUtils;
import org.coderu.common.utils.PairUtils;
import org.coderu.common.utils.types.Pair;
import org.coderu.core.impl.common.types.Package;
import org.coderu.core.impl.common.types.PackageDependency;
import org.coderu.core.impl.common.types.PackageName;
import org.coderu.core.impl.common.types.PackageType;

import com.google.common.base.Function;
import com.google.common.base.Predicate;
import com.google.common.base.Predicates;
import com.google.common.collect.Collections2;

/**
 * Generates the access rights according the common access rules:

 * 1. Each package has access to all its direct children;
 * 2. If package "P1" has access to package "P2","P1" has access to all direct children of "P2", 
 *    except "IMPL". This rule is recursive.
 * 3. Each package "P1" != "API" and all successors from "P1" have access to its sibling "COMMON" package.
 * 4. Each package "P1" has access to "COMMON" that is sibling of "P1" ancestor. 
 * 5. "FACTORY" package and all its successors have access to all siblings of "FACTORY" package;
 * 6. Each package has access to any "API" package. 
 *    Exceptions: 
 *      - "IMPL" border should not be crossed downwards.
 *      - access to ancestors is not permitted. 
 * 
 * @author arubinov
 *
 */
public final class AccessRightsFactory {

    @SuppressWarnings("unchecked")
    public static Collection<PackageDependency> generateAccessRights(final Collection<Package> nodes){

        final Collection<Package> apis = 
                getAllNodes(nodes,Package.IS_API, TraversingType.GO_ON_AT_NOT_MATCHING);

        final Collection<Package> implPackages = 
                getAllNodes(nodes,Package.IS_IMPL, TraversingType.GO_ON_AT_NOT_MATCHING);

        final Collection<Package> commonPackages = 
                getAllNodes(nodes,Package.IS_COMMON, TraversingType.GO_ON_AT_NOT_MATCHING);

        final Collection<Package> factories = 
                getAllNodes(nodes,Package.IS_FACTORY, TraversingType.GO_ON_AT_NOT_MATCHING);

        final Collection<PackageDependency> topDownRights =
                createTopDownRights(concat(nodes, implPackages));

        final Collection<PackageDependency> commonsRights =
                createRightsToCommon(commonPackages);

        final Collection<PackageDependency> factoryRights =
                createFactoryRights(factories);

        final Collection<PackageDependency> apiRights =
                createApiRights(nodes, apis);

        final Collection<PackageDependency> toSelfRights =
                createToSelfRights(nodes);

        return concat(toSelfRights, topDownRights, commonsRights, factoryRights, apiRights);

    }

    private static Collection<PackageDependency> createToSelfRights(Collection<Package> nodes) {
        final Collection<Package> allNodes = getAllNodes(nodes);
        return transform(allNodes, PACKAGE_TO_TOSELF_RIGHT);
    }

    private static Collection<PackageDependency> createApiRights(
            Collection<Package> nodes, Collection<Package> apis) {
        
        final Collection<Package> allNodes = getAllNodes(nodes);
        
        final Collection<Package> api = getAllNodes(apis, IS_NOT_IMPL, TraversingType.STOP_AT_NOT_MATCHING);
        
        return removeAccessToParentAndSelf(createRights(allNodes, api));
    }


    private static Collection<PackageDependency> createFactoryRights(Collection<Package> factoryPackages) {
        return concat(Collections2.transform(factoryPackages, FACTORY_TO_RIGHTS));
    }

    private static final Function<Package, Collection<PackageDependency>> FACTORY_TO_RIGHTS = 
            new Function<Package, Collection<PackageDependency>>() {

        public Collection<PackageDependency> apply(Package o) {
            return createFactoryRights(o);
        }
    };

    private static Collection<PackageDependency> createFactoryRights( Package factory ){
        final Collection<Package> siblings = factory.getSiblings();

        final Collection<Package> notImplChildsOfSiblings = 
                getAllNodes( 
                        concat(transform(siblings, Package.TO_CHILDS)), 
                        IS_NOT_IMPL, 
                        TraversingType.STOP_AT_NOT_MATCHING);

        final Collection<Package> to = concat(siblings, notImplChildsOfSiblings);
        
        final Collection<Package> from = getAllNodes(factory);
        
        return createRights(from, to);
    }

    private static Collection<PackageDependency> createRightsToCommon(Collection<Package> commonPackages) {
        return concat(transform(commonPackages, COMMON_TO_RIGHTS));
    }

    private static final Function<Package, Collection<PackageDependency>> COMMON_TO_RIGHTS = 
            new Function<Package, Collection<PackageDependency>>() {

        public Collection<PackageDependency> apply(Package o) {
            return createCommonRight(o);
        }
    };

    private static Collection<PackageDependency> createCommonRight( Package common ){
        final Collection<Package> siblings = Collections2.filter(common.getSiblings(), NOT_API) ;

        final Collection<Package> from = 
                getAllNodes(siblings, ALWAYS_TRUE, TraversingType.STOP_AT_NOT_MATCHING);

        final Collection<Package> to = 
                getAllNodes(common, IS_NOT_IMPL,  TraversingType.STOP_AT_NOT_MATCHING);
        
        return createRights(from, to);
    }

    private static Collection<PackageDependency> createRights(
            final Collection<Package> froms, final Collection<Package> tos) {

        final Collection<PackageDependency> result = 
                new ArrayList<PackageDependency>(froms.size()*tos.size());
        for(Package from: froms){
            for(Package to: tos){
                result.add(createPackageDependency(from.getId(),to.getId()));
            }
        }
        return result;
    }

    private static Collection<Package> getAllNodes(final Package packagge){
        return getAllNodes(packagge, ALWAYS_TRUE, TraversingType.STOP_AT_NOT_MATCHING);
    }

    private static Collection<Package> getAllNodes(final Collection<Package> packagge){
        return getAllNodes(packagge, ALWAYS_TRUE, TraversingType.STOP_AT_NOT_MATCHING);
    }

    //@Cached
    private static Collection<Package> getAllNodes(
            final Package packagge, 
            final Predicate<Package> predicate,
            final TraversingType traversingType){

        if( !predicate.apply(packagge) && traversingType == TraversingType.STOP_AT_NOT_MATCHING){ 
            return Collections.<Package>emptyList();
        }else{
            return concat( getAllNodes(packagge.getChilds(), predicate, traversingType), 
                    Collections2.filter(Collections.singletonList(packagge), predicate));
        }
    }

    //@Cached
    private static Collection<Package> getAllNodes(
            final Collection<Package> root, 
            final Predicate<Package> predicate,
            final TraversingType traversingType){
        return concat( transform( root, toAllChilds(predicate, traversingType)));
    }

    private static Function<Package, Collection<Package>> toAllChilds(
            final Predicate<Package> predicate,
            final TraversingType traversingType){

        return new Function<Package, Collection<Package>>() {

            public Collection<Package> apply(Package o) {
                return getAllNodes(o, predicate, traversingType);
            }
        };
    }

    //@Cached
    private static Collection<PackageDependency> createTopDownRights(
            final Collection<Package> nodes) {

        return getTopDownRightsAndChilds(nodes).getFirst();
    }

    private static Pair<Collection<PackageDependency>,Collection<Package>> getTopDownRightsAndChilds(
            final Collection<Package> nodes) {

        final Collection<Pair<Collection<PackageDependency>, Collection<Package>>> coll = 
            transform(nodes, package2rights());

        return Pair.make(concat(transform(coll, PAIR_TO_FIRST)), concat(transform(coll, PAIR_TO_SECOND)));
    }

    private static final Function<Package, Pair<Collection<PackageDependency>,Collection<Package>>> 
    package2rights(){
        return new Function<Package, Pair<Collection<PackageDependency>,Collection<Package>>>(){

            public Pair<Collection<PackageDependency>,Collection<Package>> apply(Package node) {

                return createTopDownRights(node);
            }

        };
    } 

    @SuppressWarnings("unchecked")
    private static Pair<Collection<PackageDependency>,Collection<Package>> createTopDownRights(
            final Package node) {

        final Collection<Package> childsWithoutImpl = 
                Collections2.filter(node.getChilds(), IS_NOT_IMPL);

        final Collection<Package> childsImpl = 
                Collections2.filter(node.getChilds(), Package.IS_IMPL);

        final Pair<Collection<PackageDependency>,Collection<Package>> rightsAndchilds = 
                getTopDownRightsAndChilds(childsWithoutImpl);

        final Collection<PackageDependency> accessToImplChilds = 
                createRigths(node, concat(childsImpl,getTopDownRightsAndChilds(childsImpl).getSecond()));

        final Collection<PackageDependency> rights = 
                concat(
                        accessToImplChilds, 
                        rightsAndchilds.getFirst(), 
                        createRigths(node, rightsAndchilds.getSecond()));

        final Collection<Package> nodes = 
                concat( rightsAndchilds.getSecond(), Collections.singleton(node));

        return Pair.make(rights, nodes);
    }

    private static Collection<PackageDependency> removeAccessToParentAndSelf(
            Collection<PackageDependency> right) {
        
        return Collections2.filter(right, NOT_ACCESS_TO_PARENT_OR_SIBLING_UNDER_API_OR_SELF);
    }

    private static final Function<Package, PackageDependency> PACKAGE_TO_TOSELF_RIGHT = 
            new Function<Package, PackageDependency>() {

        public PackageDependency apply(Package o) {
            return createPackageDependency(o.getId(), o.getId());
        }
    };

    private static ArrayList<PackageName> findApiInPath(final PackageName id){
        return findPackageNameInPath(id, IS_API);
    }

    private static ArrayList<PackageName> findApiOrCommonInPath(final PackageName id){
        return findPackageNameInPath(id, IS_API_OR_COMMON);
    }

    //@Cached
    //TODO make recursive and cached
    private static ArrayList<PackageName> findPackageNameInPath(
        final PackageName id, final Predicate<String> hasName) {

        final Collection<String> name = id.getName();
        final ArrayList<String> arr = new ArrayList<String>(name.size());
        final ArrayList<PackageName> result = new ArrayList<PackageName>();
        for(String s: name){
            arr.add(s);
            if(hasName.apply(s)){
                result.add(new PackageName(new ArrayList<String>(arr)));
            }
        }
        return result;
    }
    
    private static final Predicate<String> IS_API = isTypeOf(PackageType.API);
    private static final Predicate<String> IS_API_OR_COMMON = isTypeOf(PackageType.API, PackageType.COMMON);
    
    private static Predicate<String> isTypeOf(final PackageType... types){
        final List<PackageType> list = Arrays.asList(types);
        return new Predicate<String>(){public boolean apply(String o) {
            return CollectionUtils.exists( list, compose(equalTo(o), PackageType.TO_VALUE));
        }};
    }
    
    private static Collection<PackageDependency> createRigths(Package from, Collection<Package> to) {
        
        return transform( transform(to, Package.TO_ID),PackageDependency.bindFrom(from.getId()));
    }

    private static final Predicate<Package> ALWAYS_TRUE = alwaysTrue();
    
    private static final Predicate<Package> IS_NOT_IMPL = not(Package.IS_IMPL);

    private static final Predicate<Package> NOT_API = not(Package.IS_API);

    private static final 
    Function<Pair<Collection<PackageDependency>, Collection<Package>>,Collection<PackageDependency>> 
    PAIR_TO_FIRST = PairUtils.pair2first();

    private static final 
    Function<Pair<Collection<PackageDependency>, Collection<Package>>,Collection<Package>> 
    PAIR_TO_SECOND = PairUtils.pair2second();

    private enum TraversingType{
        STOP_AT_NOT_MATCHING,
        GO_ON_AT_NOT_MATCHING,
    }
    
    private final static Predicate<PackageDependency> ACCESS_TO_SELF = new Predicate<PackageDependency>(){

        public boolean apply(PackageDependency o) {
            return o.getTo().equals( o.getFrom() );
        }

        @Override
        public String toString() {
        
            return "access to self";
        }
    }; 

    private static final Predicate<PackageDependency> ACCESS_TO_PARENT = new Predicate<PackageDependency>(){

        public boolean apply(PackageDependency o) {
            return o.getTo().isAncestorOf_( o.getFrom() );
        }

        @Override
        public String toString() {
            return "access to parent";
        }
    }; 

    private static final Predicate<PackageDependency> ACCESS_TO_BEHIND_IMPL = 
        new Predicate<PackageDependency>() {
        
        public boolean apply(PackageDependency o) {
        
            final Iterator<String> fromIter = o.getFrom().getName().iterator();
            final Iterator<String> toIter = o.getTo().getName().iterator();
            
            while(fromIter.hasNext() && toIter.hasNext()){ 
                final String from = fromIter.next();
                final String to = toIter.next();
                if(!from.equals(to)){
                    if(to.equals(PackageType.IMPL.getValue())){
                        return true;
                    }
                    break;
                }
            }
            while(toIter.hasNext()){
                final String to = toIter.next();
                if(to.equals(PackageType.IMPL.getValue())){
                    return true;
                }
            }
                
            return false;
        }
    }; 

    private static final Predicate<PackageDependency> ACCESS_TO_SIBLING_UNDER_API = 
            new Predicate<PackageDependency>(){

        public boolean apply(PackageDependency o) {
            final ArrayList<PackageName> from = findApiInPath(o.getFrom());
            final ArrayList<PackageName> to = findApiInPath(o.getTo());

            final PackageName lastApi = to.get(to.size()-1);
            return from.contains(lastApi);
        }
        
        @Override
        public String toString() {
            return "access to sibling";
        }
    }; 

    private static Predicate<PackageName> hasApiParentAsSibling(final PackageName p){
        return new Predicate<PackageName>() {
            
            public boolean apply(PackageName o) {
            
                final String packageNameSibling2Param = p.getName().get(o.getName().size()-1);
                return packageNameSibling2Param.equals(PackageType.API.getValue());
            }
        };
    }
    
    private static final Predicate<PackageDependency> ACCESS_TO_NEPHEW_FROM_API_OR_COMMON = 
        new Predicate<PackageDependency>() {
        
        public boolean apply(PackageDependency o) {
            final ArrayList<PackageName> from = findApiOrCommonInPath(o.getFrom());
            final PackageName to = o.getTo();
            return exists(from, and(isOncleOf(to), not(hasApiParentAsSibling(to))));
        }
    };
    private static PackageDependency createPackageDependency(final PackageName from, final PackageName to){
        return new PackageDependency(from,to);
    }

    @SuppressWarnings("unchecked")
    private static final Predicate<PackageDependency> NOT_ACCESS_TO_PARENT_OR_SIBLING_UNDER_API_OR_SELF = 
        Predicates.<PackageDependency>not(Predicates.<PackageDependency>or(Arrays.asList(
        ACCESS_TO_PARENT,
        ACCESS_TO_SIBLING_UNDER_API,
        ACCESS_TO_SELF,
        ACCESS_TO_BEHIND_IMPL,
        ACCESS_TO_NEPHEW_FROM_API_OR_COMMON
        )));


}
